## Load() tests
## Compare a version of Load() with the default
path_to_reference_version <- '~/test-s2d/s2dv-ref'
path_to_test_version <- '~/test-s2d/s2dv-new'
new_args <- c('remapcells')
obsolete_args <- c()

#module load R/3.2.0-foss-2015a-bare
#module load CDO
#module load NCO
.libPaths('~/tmp')
suppressMessages(library(devtools))
suppressMessages(library(ncdf4))

## Function to be used whenever needed to specify that an argument must be empty
missing_argument <- function() quote(expr = )

## Variables needed in plot_timeseries.R
fnc <- nc_open(system.file("doc/masks", "ecearth_mm_lsm.nc", package = "s2dverification"))
mask <- ncvar_get(fnc, 'LSM')
nc_close(fnc)

fnc <- nc_open(system.file("doc/masks", "gpcc_v6_land_sea_mask.nc", package = "s2dverification"))
mask_gpcc <- ncvar_get(fnc, 'lsm')
nc_close(fnc) 

fnc <- nc_open("data/mask_cru_land.nc")
mask_cru <- ncvar_get(fnc, 'pre')
nc_close(fnc)

##configurations to test
test_batteries <- list(
  ## Parameters in each call in a test battery will accumulate and apply to next calls
  ## in the same test battery.

  ## General test battery
  list(
    # 1 
    list(
      var = 'tos',
      exp = c('i00k'),
      sdates = paste0(1990, '1101'),
      output = 'lonlat'
    ),
    list(
      sdates = paste0(1990:1993, '1101')
    ),
    list(
      grid = 't130grid'
    ),
    list(
      grid = 'r250x125'
    ),
    # 5
    list(
      grid = missing_argument(),
      lonmin = 90,
      lonmax = 210
    ),
    list(
      latmin = 0,
      latmax = 60
    ),
    list(
      leadtimemin = 3,
      leadtimemax = 6
    ),
    list(
      method = 'bilinear'
    ),
    list(
      method = 'bicubic'
    ),
    # 10
    list(
      method = 'distance-weighted',
      remapcells = 500
    ),
    list(
      method = missing_argument(),
      remapcells = missing_argument(),
      output = 'lon'
    ),
    list(
      output = 'lat'
    ),
    list(
      output = 'areave'
    ),
    list(
      output = 'lonlat',
      grid = 't130grid',
      method = 'bilinear'
    ),
    # 15
    list(
      method = 'bicubic'
    ),
    list(
      method = 'distance-weighted'
    ),
    list(
      method = 'conservative'
    ),
    list(
      grid = 't43grid',
      method = 'bilinear'
    ),
    list(
      method = 'bicubic'
    ),
    # 20
    list(
      method = 'distance-weighted'
    ),
    list(
      method = 'conservative'
    )
  ),
  ## Test battery from test_load.R
  list(
    list(
      var = 'tos',
      exp = c('cnrm-cm5dec', 'gfdlhis'),
      obs = 'ERSST',
      sdates = c('19810101', '19910101', '20010101'),
      leadtimemin = 13,
      leadtimemax = 24,
      method = 'distance-weighted',
      grid = 'r180x90',
      output = 'lonlat'
    ),
    list(
      var = 'hfssd',
      exp = 'EnsCerfacsDec',
      obs = 'ERA40',
      sdates = c('19601101', '19801101', '20051101'),
      leadtimemin = missing_argument(),
      leadtimemax = missing_argument(),
      lonmin = 120,
      lonmax = 270,
      latmin = 0,
      latmax = 65,
      method = 'bicubic',
      grid = 't42grid',
      output = 'lonlat'
    ),
    list(
      var = 'tas',
      exp = 'EnsCmccSeas',
      obs = 'ERAint',
      sdates = c('19791101', '19851101', '19911101'),
      lonmin = missing_argument(),
      lonmax = missing_argument(),
      latmin = missing_argument(),
      latmax = missing_argument(),
      method = missing_argument(),
      grid = missing_argument(),
      output = missing_argument(),
      storefreq = 'daily'
    ),
    list(
      var = 'tas',
      exp = c('ECMWF_S3Seas', 'ECMWF_S4_sea', 'ECMWF_S4_ann'),
      obs = 'HadCRUT4',
      sdates = c('19851101', '19911101'),
      nmember = 51,
      nleadtime = 13,
      storefreq = missing_argument()
    ),
    # 5
    list(
      var = 'prlr',
      exp = 'DePreSysAsimDec',
      obs = 'CRU',
      sdates = c('19651101', '19851101'),
      nmember = missing_argument(),
      nleadtime = missing_argument(),
      output = 'lonlat'
    )
  ),
  ## Test battery from plot_timeseries.R
  list(
    list(
      var = 'sieN',
      exp = 'i00k',
      obs = 'HasISST',
      sdates = paste0(seq(1960, 2005, 5), '1101'),
      leadtimemax = 60
    ),
    list(
      var = 'sieS'
    ),
    list(
      var = 'siaN'
    ),
    list(
      var = 'siaS'
    ),
    # 5
    list(
      var = 'sivN',
      obs = 'PIOMAS'
    ),
    list(
      var = 'sivS'
    ),
    list(
      var = 'tas',
      obs = c('NCEP', 'ERA40')
    ),
    list(
      var = 'heatc',
      obs = 'NEMOVAR_S4'
    ),
    list(
      var = '0-315_heatc'
    ),
    # 10
    list(
      var = '373-657_heatc'
    ),
    list(
      var = '800-5350_heatc'
    ),
    list(
      var = 'moc_40N55N_1-2km'
    ),
    list(
      var = 'tos',
      obs = c('ERSST', 'HadISST'),
      latmin = -60,
      latmax = 65,
      maskmod = list(1 - mask)
    ),
    list(
      var = 'prlr',
      obs = c('CRU', 'GPCC'),
      maskmod = list(mask),
      maskobs = list(mask_cru, mask_gpcc)
    ),
    # 15
    list(
      var = 'sieN',
      exp = 'b02p',
      obs = missing_argument(),
      sdates = '19501101',
      latmin = missing_argument(),
      latmax = missing_argument(),
      maskmod = missing_argument(),
      maskobs = missing_argument()
    ),
    list(
      var = 'sieS'
    ),
    list(
      var = 'siaN'
    ),
    list(
      var = 'siaS'
    ),
    list(
      var = 'sivN'
    ),
    # 20
    list(
      var = 'sivS'
    ),
    list(
      var = 'tas'
    ),
    list(
      var = 'heatc'
    ),
    list(
      var = '0-315_heatc'
    ),
    list(
      var = '373-657_heatc'
    ),
    # 25
    list(
      var = '800-5350_heatc'
    ),
    list(
      var = 'moc_40N55N_1-2km'
    ),
    list(
      var = 'tos',
      latmin = -60,
      latmax = 65,
      maskmod = list(1 - mask)
    ),
    list(
      var = 'prlr',
      maskmod = list(mask)
    )
  ),
  ## Test battery from plot_maps.R
  list(
    list(
      var = 'tas',
      exp = 'i00k',
      obs = 'GHCNERSSTGISS',
      sdates = paste0(seq(1960, 2005, 5), '1101'),
      leadtimemin = 61,
      leadtimemax = 108,
      output = 'lonlat'
    ),
    list(
      var = 'tos',
      obs = 'ERSST'
    ),
    list(
      var = 'prlr',
      obs = 'GPCC'
    ),
    list(
      var = 'tas',
      exp = 'b02p',
      obs = missing_argument(),
      sdates = '19501101',
      leadtimemin = missing_argument(),
      leadtimemax = missing_argument()
    )
  ),
  # Other tests
  list(
    list(
      var = 'tas',
      exp = 'EnsCmccSeas',
      obs = 'ERAint',
      sdates = '19750501',
      output = 'lonlat',
      lonmin = -20,
      lonmax = 70,
      latmin = 25,
      latmax = 75
    ),
    list(
      var = 'tos',
      exp = 'm02j',
      obs = 'HadISST',
      sdates = '19931101',
      lonmin = -60,
      lonmax = 30,
      leadtimemax = 4,
      grid = 'r300x200'
    ),
    list(
      grid = 'r360x180'
    )
  )
)

fails_dir <- './fails'
if (!dir.exists(fails_dir)) {
  dir.create(fails_dir)
}

success_dir <- './success'
if (!dir.exists(success_dir)) {
  dir.create(success_dir)  
}

for (battery_index in 1:length(test_batteries)) {
  battery <- test_batteries[[battery_index]]
  calls_to_test_indices <- 1:length(battery)
  if (!dir.exists(paste0(fails_dir, '/', battery_index))) {
    dir.create(paste0(fails_dir, '/', battery_index))
  } else {
    discard_calls <- as.numeric(gsub("[^0-9]", "", list.files(paste0(success_dir, '/', battery_index))))
    if (length(discard_calls) > 0) {
      calls_to_test_indices <- calls_to_test_indices[-discard_calls]
    }
  }
  if (!dir.exists(paste0(success_dir, '/', battery_index))) {
    dir.create(paste0(success_dir, '/', battery_index))
  }

  call <- list()
  for (call_index in 1:length(battery)) {
    cat(paste0("Battery ", battery_index, ", Test ", call_index, "...\n"))
    call[names(battery[[call_index]])] <- battery[[call_index]]

    if (call_index %in% calls_to_test_indices) {  
      suppressMessages(library(s2dverification, lib.loc = path_to_reference_version))
      error_old <- try({
        time_old <- system.time({
          sink("/dev/null")
          discard_args <- na.omit(match(new_args, names(call)))
          if (length(discard_args) > 0) {
            old_data <- do.call(Load, call[-discard_args])
          } else {
            old_data <- do.call(Load, call)
          }
          attributes(old_data$mod) <- NULL
          attributes(old_data$obs) <- NULL
          sink()
        })
      })
      print(time_old)
      unload(find.package('s2dverification'))
      suppressMessages(library(s2dverification, lib.loc = path_to_test_version))
      error_new <- try({
        time_new <- system.time({
          sink("/dev/null")
          discard_args <- na.omit(match(obsolete_args, names(call)))
          if (length(discard_args) > 0) {
            new_data <- do.call(Load, call[-discard_args])
          } else {
            new_data <- do.call(Load, call)
          }
          attributes(new_data$mod) <- NULL
          attributes(new_data$obs) <- NULL
          attributes(new_data$lon) <- NULL
          attributes(new_data$lat) <- NULL
          sink()
        })
      })
      print(time_new)
      unload(find.package('s2dverification'))
      fail <- NULL
      if ('try-error' %in% class(error_old) ||
          'try-error' %in% class(error_new)) {
        fail <- 'Load crashed'
      } else {
        ##tests to perform
        if (!(identical(old_data$mod, new_data$mod) &&
              identical(old_data$obs, new_data$obs))) { ## &&
              #identical(old_data$lon, new_data$lon))) { ## &&
              #identical(old_data$lat, new_data$lat))) {
          fail <- 'Unequal data'
        }
        if (length(old_data$lon) != length(new_data$lon)) {
          fail <- c(fail, 'Different numbers of lons.')
        }
        if (length(old_data$lat) != length(new_data$lat)) {
          fail <- c(fail, 'Different numbers of lats.')
        }
        if (!identical(as.vector(signif(old_data$lon, 4)), signif(new_data$lon, 4))) {
          fail <- c(fail, 'Lons not equal signif 4.')
        }
        if (!identical(as.vector(signif(old_data$lat, 4)), signif(new_data$lat, 4))) {
          fail <- c(fail, 'Lats not equal signif 4.')
        }
      }
    
      if (is.null(fail)) {
        if (file.exists(paste0(fails_dir, '/', battery_index, '/', call_index, '.RData'))) {
          file.remove(paste0(fails_dir, '/', battery_index, '/', call_index, '.RData'))
        }
        writeLines('OK', paste0(success_dir, '/', battery_index, '/', call_index))
        cat('OK\n')
      } else {
        cat(paste0('FAIL (', paste(fail, collapse = ' // '), ')\n'))
        save(call, old_data, new_data, file = paste0(fails_dir, '/', battery_index, '/', call_index, '.RData'))
      }
    
      rm(old_data, new_data)
    }
  }
}
